from collections import defaultdict

class GraphNode(object):

    def __init__(self, time, obj):
        self.time = time
        self.obj = obj


class GraphStructure(object):

    def __init__(self, connections=None):
        """
        Initialize graph
        """
        self._graph = defaultdict(set)

        if connections is not None:
            self.add_connections(connections)

    """
    def add_nodes(self, nodes):
        for node in nodes:
            if node not in self._graph:
                self._graph[node] = set()
    """

    def add_connections(self, connections):
        """ Add connections (pb2.Connections) to graph """

        for con in connections:
            self.add(con.n1, con.n2)

    def add(self, node1, node2):
        """ Add connection between node1 and node2 (pb ObjectNodes) """

        self._graph[node1].add(node2)
        if node2 not in self._graph:
            # print("add empty node")
            self._graph[node2] = set()
        else:
            pass
            # print("already exist")

    def remove(self, node):
        """ Remove all references to node """

        for n, cxns in self._graph.items():  # python3: items(); python2: iteritems()
            try:
                cxns.remove(node)
            except KeyError:
                pass
        try:
            del self._graph[node]
        except KeyError:
            pass

    def is_connected(self, node1, node2):
        """ Is node1 directly connected to node2 """

        return node1 in self._graph and node2 in self._graph[node1]

    def find_path(self, node1, node2, path=[]):
        """ Find any path between node1 and node2 (may not be shortest) """

        path = path + [node1]
        if node1 == node2:
            return path
        if node1 not in self._graph:
            return None
        for node in self._graph[node1]:
            if node not in path:
                new_path = self.find_path(node, node2, path)
                if new_path:
                    return new_path
        return None

    @staticmethod
    def get_downstream_node_indices(graph_list, start_idx):
        node_indices = [start_idx]

        node, connected_nodes = graph_list[start_idx]
        for c_node in connected_nodes:
            # get index of connected node in graph
            c_node_idx = [n_ for n_, cv_ in graph_list].index(c_node)
            # call recursively on this connected node
            c_node_downstream_indices = WaveGraph.get_downstream_node_indices(graph_list, c_node_idx)
            node_indices.extend(c_node_downstream_indices)

        return list(set(node_indices))

    def get_items(self):
        return self._graph.items()

    # extract wave objects from graph: question here: what is a wave?
    #   e.g. before/after merging, what belongs to same wave
    # wave object is list of wavetroughs, so wavetroughs for consecutive timesteps
    def extract_waves(self):

        # sort nodes by time (is in key), as list here
        time_sorted_nodes = list(sorted(self._graph.items(), key=lambda item: item[0].time))

        print(time_sorted_nodes)

        wave_id_per_trough = [None] * len(time_sorted_nodes)
        cur_id = 0

        # iterate over all time sorted identified troughs
        # search temporal downstream tracked troughs and group them using a set id
        for trough_idx, (trough_node, connected_nodes) in enumerate(time_sorted_nodes):

            if wave_id_per_trough[trough_idx] is not None:  # already part of a wave
                continue

            # not part of wave -> get (temporal) downstream connections = wave (return indices of them)
            downstream_wave_node_indices = WaveGraph.get_downstream_node_indices(time_sorted_nodes, trough_idx)

            # any of downstream nodes already part of a wave?
            connected_wave_id = None
            for ds_node_idx in downstream_wave_node_indices:
                if wave_id_per_trough[ds_node_idx] is not None:
                    if connected_wave_id is not None:
                        print("Double ID...")
                    connected_wave_id = wave_id_per_trough[ds_node_idx]

            # if so set all nodes to this found id
            if connected_wave_id is not None:
                for ds_node_idx in downstream_wave_node_indices:
                    wave_id_per_trough[ds_node_idx] = connected_wave_id
                continue

            # else new path for all wave_nodes
            cur_id_needs_update = False
            for ds_node_idx in downstream_wave_node_indices:
                wave_id_per_trough[ds_node_idx] = cur_id
                cur_id_needs_update = True

            if cur_id_needs_update:
                cur_id += 1
        print(wave_id_per_trough)

        # done, now extract every wave by id and put them into subgraphs

        waves = []  # generate list of waves according to above policy

        for wave_id in range(cur_id):
            trough_idxs = [i for i in range(len(wave_id_per_trough)) if wave_id_per_trough[i] == wave_id]

            cur_troughs = [time_sorted_nodes[i] for i in trough_idxs]  # troughs of this wave
            # cur_troughs = cur_troughs.sortbytime # already sorted?

            """
            cur_time = cur_troughs[0].time
            for trough in cur_troughs:
                trough_list = [] #
                if trough.time == cur_time:
                    trough_list.append(trough)
                    continue

                cur_time = trough.time # next time step
                list_of_trough_states.append(WaveTroughState(-1, -1, trough_list)) # TODO
                wavetroughstate = None # reset to new one

            wave = Wave(list_of_trough_states)
            waves.append(wave)
            """
            sub_wg = WaveGraph(graph_items=cur_troughs)  # TODO cur_troughs as defaultdict(set)
            waves.append(sub_wg)
        print("Done")
        return waves

    def __str__(self):
        return '{}({})'.format(self.__class__.__name__, dict(self._graph))

